package graph;

import java.awt.*;

public class GraphPanel
        extends MouseManipulations implements Runnable {

    private ControlPanel graph;
    private Thread relaxer;

    GraphPanel(ControlPanel _graph) {
        graph = _graph;
        addMouseListener(this);
    }

    public void run() {
        relaxer(Thread.currentThread());
    }

    private void relaxer(Thread me) {
        while (relaxer == me) {
            relax();
            if (isRandom() && (Math.random() < 0.03)) {
                Node n = nodes[
                        (int) (Math.random() * nnodes)];
                if (!n.isFixed())
                    nIsNotFixed(n);
            }
            try {
                Thread.sleep(10);
            } catch (InterruptedException e) {
                break;
            }
        }
    }

    private void nIsNotFixed(Node n) {
        n.setX(n.getX() + (100 * Math.random() - 50));
        n.setY(n.getY() + (100 * Math.random() - 50));
    }

    synchronized void relax() {
        setupEdges();
        for (int i = 0; i < nnodes; i++)
            preturbJRelativetoI(i, nodes[i]);
        for (int i = 0; i < nnodes; i++)
            moveNode(nodes[i]);
        repaint();
    }

    private void preturbJRelativetoI(int i,
                                     Node n1) {
        double dx = 0;
        double dy = 0;
        for (int j = 0; j < nnodes; j++) {
            if (i == j) continue;
            Node n2 = nodes[j];
            double vx = n1.getX() - n2.getX();
            double vy = n1.getY() - n2.getY();
            double len = vx * vx + vy * vy;
            if (len == 0) {
                dx += Math.random();
                dy += Math.random();
            } else if (len < 100 * 100) {
                dx += vx / len;
                dy += vy / len;
            }
        }
        double dlen = dx * dx + dy * dy;
        if (dlen > 0) {
            dlen = Math.sqrt(dlen) / 2;
            n1.setDx(n1.getDx() + dx / dlen);
            n1.setDy(n1.getDy() + dy / dlen);
        }
    }

    /**
     * moveNode nuges the node if it is not fixed
     * it also checks the extremes and
     * reduces the dimensions
     */
    private void moveNode(Node node) {
        if (!node.isFixed())
            node.nuge();
        node.ceilingNode(getSize());
        node.oneHalfDxDy();
    }


    public void start() {
        relaxer = new Thread(this);
        relaxer.start();
    }

    public void paintNode(Graphics g, Node n, FontMetrics fm) {
        int x = (int) n.getX();
        int y = (int) n.getY();
        g.setColor((n == pick)
                ? getSelectColor() : (n.isFixed()
                ? getFixedColor() : getNodeColor()));
        int w = fm.stringWidth(n.getLbl()) + 10;
        int h = fm.getHeight() + 4;
        g.fillRect(x - w / 2, y - h / 2, w, h);
        g.setColor(Color.black);
        g.drawRect(x - w / 2, y - h / 2, w - 1, h - 1);
        g.drawString(n.getLbl(), x - (w - 10) / 2, (y - (h - 4) / 2) + fm.getAscent());
    }

    public synchronized void update(Graphics g) {
        Dimension d = getSize();
        if ((offscreen == null) ||
                (d.width != offscreensize.width) ||
                (d.height != offscreensize.height))
            createBuffer(getSize());

        offgraphics.setColor(getBackground());
        offgraphics.fillRect(0, 0, d.width, d.height);
        for (int i = 0; i < nedges; i++)
            drawArcAndString(i);
        drawNodes();
        g.drawImage(offscreen, 0, 0, null);
    }

    private void drawNodes() {
        FontMetrics fm = offgraphics.getFontMetrics();
        for (int i = 0; i < nnodes; i++)
            paintNode(offgraphics, nodes[i], fm);
    }

    private void drawArcAndString(int i) {
        Edge e = edges[i];
        int x1 = (int) nodes[e.getFrom()].getX();
        int y1 = (int) nodes[e.getFrom()].getY();
        int x2 = (int) nodes[e.getTo()].getX();
        int y2 = (int) nodes[e.getTo()].getY();
        int len = (int) Math.abs(Math.sqrt((x1 - x2) * (x1 - x2) + (y1 - y2) * (y1 - y2)) - e.getLen());
        offgraphics.setColor((len < 10) ? getArcColor1() : (len < 20 ? getArcColor2() : getArcColor3()));
        offgraphics.drawLine(x1, y1, x2, y2);
        if (isStress())
            drawArcString(len, x1, x2, y1, y2);
    }

    private void drawArcString(int len, int x1, int x2, int y1, int y2) {
        String lbl = String.valueOf(len);
        offgraphics.setColor(getStressColor());
        offgraphics.drawString(lbl, x1 + (x2 - x1) / 2, y1 + (y2 - y1) / 2);
        offgraphics.setColor(getEdgeColor());
    }

}
